package dev.fitko.fitconnect.core.routing;

import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.route.AreaResult;
import dev.fitko.fitconnect.api.domain.model.route.RouteResult;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import dev.fitko.fitconnect.api.services.routing.RoutingService;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.UriComponentsBuilder;

import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.List;
import java.util.function.Predicate;

import static java.util.function.Predicate.not;

public class RoutingApiService implements RoutingService {

    private final RestTemplate restTemplate;
    private final ApplicationConfig config;

    public RoutingApiService(final ApplicationConfig config, final RestTemplate restTemplate) {
        this.config = config;
        this.restTemplate = restTemplate;
    }

    @Override
    public AreaResult getAreas(final List<String> searchExpressions, final int offset, final int limit) throws RestApiException {

        final String url = config.getAreaEndpoint();
        final HttpHeaders headers = getHttpHeaders(MediaType.parseMediaType("application/problem+json"));
        final HttpEntity<String> entity = new HttpEntity<>(headers);

        final UriComponentsBuilder uriBuilder = UriComponentsBuilder.fromUriString(url);
        searchExpressions.forEach(search -> uriBuilder.queryParam("areaSearchexpression", search));
        uriBuilder.queryParam("offset", offset);
        uriBuilder.queryParam("limit", limit);

        try {
            return restTemplate.exchange(uriBuilder.toUriString(), HttpMethod.GET, entity, AreaResult.class).getBody();
        } catch (final RestClientException e) {
            throw new RestApiException("Area query failed", e);
        }
    }

    @Override
    public RouteResult getRoutes(final String leikaKey, final String ars, final String ags, final String areaId, final int offset, final int limit) throws RestApiException {

        testIfSearchCriterionIsSet(ars, ags, areaId);

        final String url = config.getRoutesEndpoint();
        final HttpHeaders headers = getHttpHeaders(MediaType.APPLICATION_JSON);
        final HttpEntity<String> entity = new HttpEntity<>(headers);

        final UriComponentsBuilder uriBuilder = UriComponentsBuilder.fromUriString(url);
        uriBuilder.queryParam("leikaKey", leikaKey);

        if (ars != null) {
            uriBuilder.queryParam("ars", ars);
        }
        if (ags != null) {
            uriBuilder.queryParam("ags", ags);
        }
        if (areaId != null) {
            uriBuilder.queryParam("areaId", areaId);
        }

        uriBuilder.queryParam("offset", offset);
        uriBuilder.queryParam("limit", limit);

        try {
            return restTemplate.exchange(uriBuilder.toUriString(), HttpMethod.GET, entity, RouteResult.class).getBody();
        } catch (final RestClientException e) {
            throw new RestApiException("Route query failed", e);
        }
    }

    private static void testIfSearchCriterionIsSet(final String ars, final String ags, final String areaId) {
        final List<String> areaSearchCriteria = Arrays.asList(ars, ags, areaId);

        if (areaSearchCriteria.stream().allMatch(CriterionEmpty())) {
            throw new RestApiException("At least one search criterion out of ags, ars or areaId must be set");
        }

        if (areaSearchCriteria.stream().filter(not(CriterionEmpty())).count() > 1) {
            throw new RestApiException("Only one of ars, ags or areaId must be specified.");
        }
    }

    private static Predicate<String> CriterionEmpty() {
        return criterion -> criterion == null || criterion.isEmpty();
    }

    private HttpHeaders getHttpHeaders(final MediaType mediaType) {
        final var headers = new HttpHeaders();
        headers.setAccept(List.of(mediaType));
        headers.setAcceptCharset(List.of(StandardCharsets.UTF_8));
        return headers;
    }
}
