package dev.fitko.fitconnect.core.routing;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.domain.model.route.Area;
import dev.fitko.fitconnect.api.domain.model.route.AreaResult;
import dev.fitko.fitconnect.api.domain.model.route.Route;
import dev.fitko.fitconnect.api.domain.model.route.RouteResult;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import dev.fitko.fitconnect.api.services.routing.RoutingService;
import dev.fitko.fitconnect.core.RestEndpointBase;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestTemplate;

import java.util.List;
import java.util.Map;
import java.util.UUID;

import static com.github.tomakehurst.wiremock.client.WireMock.get;
import static com.github.tomakehurst.wiremock.client.WireMock.ok;
import static com.github.tomakehurst.wiremock.client.WireMock.urlEqualTo;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;

public class RoutingApiServiceTest extends RestEndpointBase {

    private RoutingService underTest;

    @BeforeEach
    void setup() {

        final var fakeBaseUrl = "http://localhost:" + wireMockServer.port();

        final EnvironmentName envName = new EnvironmentName("TESTING");
        final Environment environment = new Environment();
        environment.setRoutingBaseUrl(fakeBaseUrl);

        final ApplicationConfig config = ApplicationConfig.builder()
                .environments(Map.of(envName, environment))
                .activeEnvironment(envName)
                .build();

        underTest = new RoutingApiService(config, new RestTemplate());
    }

    @Test
    void testFindAreas() throws JsonProcessingException {

        // Given
        final Area area = new Area();
        area.setName("Berlin");

        final AreaResult expectedResult = new AreaResult();
        expectedResult.setCount(1);
        expectedResult.setOffset(0);
        expectedResult.setTotalCount(1);
        expectedResult.setAreas(List.of(area));

        wireMockServer.stubFor(
                get(urlEqualTo("/v1/areas?areaSearchexpression=Berlin&offset=0&limit=10"))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedResult))
                                .withHeader("Content-Type", "application/problem+json")
                                .withStatus(200)));

        // When
        final AreaResult areaResult = underTest.getAreas(List.of("Berlin"), 0, 10);


        // Then
        assertNotNull(areaResult);
        assertFalse(areaResult.getAreas().isEmpty());
        assertThat(areaResult.getAreas().get(0), is(area));
    }

    @Test
    void testFindRoutes() throws JsonProcessingException {

        // Given
        final Route route = new Route();
        route.setDestinationId(UUID.randomUUID());
        route.setDestinationName("test destination");

        final RouteResult expectedResult = new RouteResult();
        expectedResult.setCount(1);
        expectedResult.setOffset(0);
        expectedResult.setTotalCount(1);
        expectedResult.setRoutes(List.of(route));

        wireMockServer.stubFor(
                get(urlEqualTo("/v1/routes?leikaKey=99123456760610&ars=064350014014&offset=0&limit=10"))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedResult))
                                .withHeader("Content-Type", MediaType.APPLICATION_JSON_VALUE)
                                .withStatus(200)));

        // When
        final RouteResult routeResult = underTest.getRoutes("99123456760610", "064350014014", null, null, 0, 10);

        // Then
        assertNotNull(routeResult);
        assertFalse(routeResult.getRoutes().isEmpty());
        assertThat(routeResult.getRoutes().get(0), is(route));
    }

    @Test
    void testFindRoutesWithNoSearchCriterion() {

        // Given
        final var leikaKey = "99123456760610";

        // When
        final RestApiException exception = assertThrows(RestApiException.class, () -> underTest.getRoutes(leikaKey, null, null, null, 0, 10));

        // Then
        assertThat(exception.getMessage(), containsString(" one search criterion out of ags, ars or areaId must be set"));

    }

    @Test
    void testFindRoutesWithMoreThanOneAreaCriterion() {

        // Given
        final var leikaKey = "99123456760610";

        // When
        final RestApiException exception = assertThrows(RestApiException.class, () -> underTest.getRoutes(leikaKey, "123", "456", null, 0, 10));

        // Then
        assertThat(exception.getMessage(), containsString("Only one of ars, ags or areaId must be specified."));

    }
}
